@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
 /* gtkhex.c - a GtkHex widget, modified for use in GHex
 
-   Copyright (C) 1998, 1999, 2000  Free Software Foundation
+   Copyright (C) 1998 - 2002 Free Software Foundation
 
    GHex is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -27,61 +27,105 @@
 #include "hex-document.h"
 #include "gtkhex.h"
 
+#include "ghex-marshal.h"
+
 #define DISPLAY_BORDER 4
 
 #define DEFAULT_CPL 32
 #define DEFAULT_LINES 16
 
-#define DEFAULT_FONT "-*-courier-medium-r-normal--12-*-*-*-*-*-*-*"
-
 #define SCROLL_TIMEOUT 100
 
-#define is_displayable(c) ((char_widths[(guchar)c])?1:0)
+/* jaKa: this causes trouble to pango */
+#if 0 
+#  define is_displayable(c) ((char_widths[(guchar)c])?1:0)
+#else 
+#  define is_displayable(c) (((c) >= 0x20) && ((c) <= 0x7f))
+#endif /* 0 */
+
+static GdkColor cl_cursor_a = { 0, 0xFFFF, 0, 0 };
+static GdkColor cl_cursor_p = { 0, 0xFFFF, 0x7F00, 0x7F00 };
 
 typedef void (*DataChangedSignal)(GtkObject *, gpointer, gpointer);
 
 enum {
 	CURSOR_MOVED_SIGNAL,
 	DATA_CHANGED_SIGNAL,
+	CUT_CLIPBOARD_SIGNAL,
+	COPY_CLIPBOARD_SIGNAL,
+	PASTE_CLIPBOARD_SIGNAL,
 	LAST_SIGNAL
 };
 
-static gint gtkhex_signals[LAST_SIGNAL] = { 0, 0 };
+enum {
+  TARGET_STRING,
+};
 
-GtkFixedClass *parent_class = NULL;
-gchar *char_widths = NULL;
+static gint gtkhex_signals[LAST_SIGNAL] = { 0 };
+
+static GtkFixedClass *parent_class = NULL;
+static gchar *char_widths = NULL;
 
 static void render_hex_lines(GtkHex *, gint, gint);
 static void render_ascii_lines(GtkHex *, gint, gint);
+static void gtk_hex_set_selection(GtkHex *gh, gint start, gint end);
 
-static void gtk_hex_marshaller (GtkObject *object, GtkSignalFunc func,
-								gpointer func_data, GtkArg *args) {
-	DataChangedSignal rfunc;
+static guint32 get_event_time() {
+	GdkEvent *event;
+	guint32 tm = GDK_CURRENT_TIME;
   
-	rfunc = (DataChangedSignal) func;
+	event = gtk_get_current_event();
   
-	(* rfunc)(object, GTK_VALUE_POINTER (args[0]), func_data);
+	if(event) {
+		switch (event->type) {
+		case GDK_MOTION_NOTIFY:
+			tm = event->motion.time; break;
+		case GDK_BUTTON_PRESS:
+		case GDK_2BUTTON_PRESS:
+		case GDK_3BUTTON_PRESS:
+		case GDK_BUTTON_RELEASE:
+			tm = event->button.time; break;
+		case GDK_KEY_PRESS:
+		case GDK_KEY_RELEASE:
+			tm = event->key.time; break;
+		case GDK_ENTER_NOTIFY:
+		case GDK_LEAVE_NOTIFY:
+			tm = event->crossing.time; break;
+		case GDK_PROPERTY_NOTIFY:
+			tm = event->property.time; break;
+		case GDK_SELECTION_CLEAR:
+		case GDK_SELECTION_REQUEST:
+		case GDK_SELECTION_NOTIFY:
+			tm = event->selection.time; break;
+		case GDK_PROXIMITY_IN:
+		case GDK_PROXIMITY_OUT:
+			tm = event->proximity.time; break;
+		default:			/* use current time */
+			break;
+		}
+		
+		gdk_event_free(event);
+	}
+  
+	return tm;
 }
 
 /*
  * simply forces widget w to redraw itself
  */
 static void redraw_widget(GtkWidget *w) {
-	GdkRectangle rect;
-	
-	rect.x = rect.y = 0;
-	rect.width = w->allocation.width;
-	rect.height = w->allocation.height;
-	
-	gtk_widget_draw(w, &rect);
+	if(!GTK_WIDGET_REALIZED(w))
+		return;
+
+	gdk_window_invalidate_rect (w->window, NULL, FALSE);
 }
 
 static gint widget_get_xt(GtkWidget *w) {
-	return w->style->klass->xthickness;
+	return w->style->xthickness; /* Changed in Gtk-2.0 -- SnM */
 }
 
 static gint widget_get_yt(GtkWidget *w) {
-	return w->style->klass->ythickness;
+	return w->style->ythickness; /* Changed in Gtk-2.0 -- SnM */
 }
 /*
  * ?_to_pointer translates mouse coordinates in hex/ascii view
@@ -118,32 +162,42 @@ static void ascii_to_pointer(GtkHex *gh, gint mx, gint my) {
 	gtk_hex_set_cursor_xy(gh, mx/gh->char_width, cy);
 }
 
-static guint get_max_char_width(GdkFont *font) {
+static guint get_max_char_width(GtkHex *gh, PangoFontMetrics *font_metrics) {
 	/* this is, I guess, a rather dirty trick, but
 	   right now i can't think of anything better */
 	guint i;
 	guint maxwidth = 0;
+	PangoRectangle logical_rect;
+	PangoLayout *layout;
+	gchar str[2]; 
 
 	if (char_widths == NULL)
 		char_widths = (gchar*)g_malloc(0x100);
 
 	char_widths[0] = 0;
-	for(i = 1; i < 0x100; i++)
-		char_widths[i] = gdk_char_width(font, (gchar)i);
-	
+
+	layout = gtk_widget_create_pango_layout (GTK_WIDGET (gh), "");
+	pango_layout_set_font_description (layout, gh->font_desc);
+
+	for(i = 1; i < 0x100; i++) {
+		logical_rect.width = 0;
+		/* Check if the char is displayable. Caused trouble to pango */
+		if (is_displayable((guchar)i)) {
+			sprintf (str, "%c", (gchar)i);
+			pango_layout_set_text(layout, str, -1);
+			pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
+		}
+		char_widths[i] = logical_rect.width;
+	}
+
 	for(i = '0'; i <= 'z'; i++)
 		maxwidth = MAX(maxwidth, char_widths[i]);
 
-/*	for(i = 0; i < 0x100; i++)
-		if(char_widths[i] && (char_widths[i] != maxwidth)) {
-			gnome_app_warning(mdi->active_window, _("This does not appear to be a monospaced font. Ghex's display may not be perfect with this font."));
-			break;
-		}*/
-	
+	g_object_unref (G_OBJECT (layout));
 	return maxwidth;
 }
 
-static void format_xbyte(GtkHex *gh, gint pos, gchar buf[2]) {
+void format_xbyte(GtkHex *gh, gint pos, gchar buf[2]) {
 	guint low, high;
 	guchar c;
 
@@ -159,7 +213,7 @@ static void format_xbyte(GtkHex *gh, gint pos, gchar buf[2]) {
  * format_[x|a]block() formats contents of the buffer
  * into displayable text in hex or ascii, respectively
  */
-static gint format_xblock(GtkHex *gh, gchar *out, guint start, guint end) {
+gint format_xblock(GtkHex *gh, gchar *out, guint start, guint end) {
 	int i, j, low, high;
 	guchar c;
 
@@ -178,7 +232,7 @@ static gint format_xblock(GtkHex *gh, gchar *out, guint start, guint end) {
 	return j;
 }
 
-static gint format_ablock(GtkHex *gh, gchar *out, guint start, guint end) {
+gint format_ablock(GtkHex *gh, gchar *out, guint start, guint end) {
 	int i, j;
 	guchar c;
 
@@ -263,8 +317,10 @@ static void render_byte(GtkHex *gh, gint pos) {
 
 	if(pos < gh->document->file_size) {
 		gdk_gc_set_foreground(gh->xdisp_gc, &GTK_WIDGET(gh)->style->text[GTK_STATE_NORMAL]);
-		gdk_draw_text(gh->xdisp->window, gh->disp_font, gh->xdisp_gc,
-					  cx, cy + gh->disp_font->ascent, buf, 2);
+		/* Changes for Gnome 2.0 */
+		pango_layout_set_text (gh->xlayout, buf, 2);
+		gdk_draw_layout (gh->xdisp->window, gh->xdisp_gc, cx, cy, gh->xlayout);
+
 	}
 	
 	if(!get_acoords(gh, pos, &cx, &cy))
@@ -276,10 +332,12 @@ static void render_byte(GtkHex *gh, gint pos) {
 	if(pos < gh->document->file_size) {
 		gdk_gc_set_foreground(gh->adisp_gc, &GTK_WIDGET(gh)->style->text[GTK_STATE_NORMAL]);
 		buf[0] = gtk_hex_get_byte(gh, pos);
-		if(!is_displayable(buf[0]))
+		if(!is_displayable((guchar)buf[0]))
 			buf[0] = '.';
-		gdk_draw_text(gh->adisp->window, gh->disp_font, gh->adisp_gc,
-					  cx, cy + gh->disp_font->ascent, buf, 1);
+		/* Changes for Gnome 2.0 */
+		pango_layout_set_text (gh->alayout, buf, 1);
+		gdk_draw_layout (gh->adisp->window, gh->adisp_gc, cx, cy, gh->alayout);
+
 	}
 }
 
@@ -289,9 +347,8 @@ static void render_byte(GtkHex *gh, gint pos) {
 static void render_ac(GtkHex *gh) {
 	gint cx, cy;
 	static guchar c[2] = "\0\0";
-	GtkStateType state;
-	GtkShadowType shadow;
-
+	GtkHexClass *klass = GTK_HEX_CLASS(GTK_WIDGET_GET_CLASS(gh));
+	
 	if(!GTK_WIDGET_REALIZED(gh->adisp))
 		return;
 	
@@ -301,27 +358,25 @@ static void render_ac(GtkHex *gh) {
 			c[0] = '.';
 
 		if(gh->active_view == VIEW_ASCII) {
-			state = GTK_STATE_SELECTED;
-			shadow = GTK_SHADOW_NONE;
+			gdk_gc_set_foreground(gh->adisp_gc, &klass->cl_cursor_a);
 		}
 		else {
-			state = GTK_STATE_NORMAL;
-			shadow = GTK_SHADOW_IN;
+			gdk_gc_set_foreground(gh->adisp_gc, &klass->cl_cursor_p);
 		}
 
-		gtk_draw_box(GTK_WIDGET(gh)->style, gh->adisp->window,
-					 state, shadow,
-					 cx, cy, gh->char_width, gh->char_height - 1);
-		gdk_draw_text(gh->adisp->window, gh->disp_font, gh->adisp_gc,
-					  cx, cy + gh->disp_font->ascent, c, 1);
+		gdk_draw_rectangle(gh->adisp->window, gh->adisp_gc,
+						   TRUE, cx, cy, gh->char_width, gh->char_height - 1);
+		gdk_gc_set_foreground(gh->adisp_gc, &(GTK_WIDGET(gh)->style->black));
+		/* Changes for Gnome 2.0 */
+		pango_layout_set_text (gh->alayout, c, 1);
+		gdk_draw_layout (gh->adisp->window, gh->adisp_gc, cx, cy, gh->alayout);
 	}
 }
 
 static void render_xc(GtkHex *gh) {
 	gint cx, cy, i;
 	static guchar c[2];
-	GtkStateType state;
-	GtkShadowType shadow;
+	GtkHexClass *klass = GTK_HEX_CLASS(GTK_WIDGET_GET_CLASS(gh));
 
 	if(!GTK_WIDGET_REALIZED(gh->xdisp))
 		return;
@@ -338,19 +393,17 @@ static void render_xc(GtkHex *gh) {
 		}
 
 		if(gh->active_view == VIEW_HEX) {
-			state = GTK_STATE_SELECTED;
-			shadow = GTK_SHADOW_NONE;
+			gdk_gc_set_foreground(gh->xdisp_gc, &klass->cl_cursor_a);
 		}
 		else {
-			state = GTK_STATE_NORMAL;
-			shadow = GTK_SHADOW_IN;
+			gdk_gc_set_foreground(gh->xdisp_gc, &klass->cl_cursor_p);
 		}
 
-		gtk_draw_box(GTK_WIDGET(gh)->style, gh->xdisp->window,
-					 state, shadow,
-					 cx, cy, gh->char_width, gh->char_height - 1);
-		gdk_draw_text(gh->xdisp->window, gh->disp_font, gh->xdisp_gc,
-					  cx, cy + gh->disp_font->ascent, &c[i], 1);
+		gdk_draw_rectangle(GTK_WIDGET(gh->xdisp)->window, gh->xdisp_gc,
+						   TRUE, cx, cy, gh->char_width, gh->char_height - 1);
+		gdk_gc_set_foreground(gh->xdisp_gc, &(GTK_WIDGET(gh)->style->black));
+		pango_layout_set_text (gh->xlayout, &c[i], 1);
+		gdk_draw_layout (gh->xdisp->window, gh->xdisp_gc, cx, cy, gh->xlayout);
 	}
 }
 
@@ -381,15 +434,27 @@ static void hide_cursor(GtkHex *gh) {
  */
 static void render_hex_lines(GtkHex *gh, gint imin, gint imax) {
 	GtkWidget *w = gh->xdisp;
-	gint i, cursor_line;
+	gint i, cursor_line, sl, el, start, end;
 	gint xcpl = gh->cpl*2 + gh->cpl/gh->group_type;
-	gint frm_len, tmp;
+	gint frm_len, tmp, cursor_off = 0, len;
 
 	if( (!GTK_WIDGET_REALIZED(gh)) || (gh->cpl == 0) )
 		return;
 
 	cursor_line = gh->cursor_pos / gh->cpl - gh->top_line;
 
+	tmp = gh->sel_start;
+	end = gh->sel_end;
+	if(tmp < end)
+		start = tmp;
+	else {
+		start = end;
+		end = tmp;
+	}
+	end--;
+	sl = start/gh->cpl - gh->top_line;
+	el = end/gh->cpl - gh->top_line;
+
 	gdk_gc_set_foreground(gh->xdisp_gc, &GTK_WIDGET(gh)->style->base[GTK_STATE_NORMAL]);
 	gdk_draw_rectangle(w->window, gh->xdisp_gc, TRUE,
 					   0, imin*gh->char_height, w->allocation.width,
@@ -407,10 +472,39 @@ static void render_hex_lines(GtkHex *gh, gint imin, gint imax) {
 		tmp = (gint)frm_len - (gint)((i - imin)*xcpl);
 		if(tmp <= 0)
 			return;
-		gdk_draw_text(w->window, gh->disp_font, gh->xdisp_gc,
-					  0, gh->disp_font->ascent + i*gh->char_height,
-					  gh->disp_buffer + (i - imin)*xcpl, 
-					  MIN(xcpl, tmp));
+
+		if(gh->sel_start != gh->sel_end) {
+			if(i == sl) {
+				cursor_off = 2*(start%gh->cpl) + (start%gh->cpl)/gh->group_type;
+				if(i == el)
+					len = 2*(end%gh->cpl + 1) + (end%gh->cpl)/gh->group_type;
+				else
+					len = xcpl;
+				len = len - cursor_off;
+				if(len > 0)
+					gtk_draw_box(GTK_WIDGET(gh)->style, gh->xdisp->window,
+								 GTK_STATE_SELECTED, GTK_SHADOW_NONE,
+								 cursor_off*gh->char_width, i*gh->char_height,
+								 len*gh->char_width, gh->char_height);
+			}
+			else if(i == el) {
+				cursor_off = 2*(end%gh->cpl + 1) + (end%gh->cpl)/gh->group_type;
+				if(cursor_off > 0)
+					gtk_draw_box(GTK_WIDGET(gh)->style, gh->xdisp->window,
+								 GTK_STATE_SELECTED, GTK_SHADOW_NONE,
+								 0, i*gh->char_height,
+								 cursor_off*gh->char_width, gh->char_height);
+			}
+			else if(i > sl && i < el) {
+				gtk_draw_box(GTK_WIDGET(gh)->style, gh->xdisp->window,
+							 GTK_STATE_SELECTED, GTK_SHADOW_NONE,
+							 0, i*gh->char_height,
+							 xcpl*gh->char_width, gh->char_height);
+			}
+		} 
+		/* Changes for Gnome 2.0 */
+		pango_layout_set_text (gh->xlayout, gh->disp_buffer + (i - imin)*xcpl, MIN(xcpl, tmp));
+		gdk_draw_layout (w->window, gh->xdisp_gc, 0, i*gh->char_height, gh->xlayout);
 	}
 	
 	if((cursor_line >= imin) && (cursor_line <= imax) && (gh->cursor_shown))
@@ -419,14 +513,26 @@ static void render_hex_lines(GtkHex *gh, gint imin, gint imax) {
 
 static void render_ascii_lines(GtkHex *gh, gint imin, gint imax) {
 	GtkWidget *w = gh->adisp;
-	gint i, tmp, frm_len;
-	guint cursor_line;
+	gint i, tmp, frm_len, sl, el, start, end;
+	guint cursor_line, cursor_off, len;
 	
 	if( (!GTK_WIDGET_REALIZED(gh)) || (gh->cpl == 0) )
 		return;
 	
 	cursor_line = gh->cursor_pos / gh->cpl - gh->top_line;
 	
+	tmp = gh->sel_start;
+	end = gh->sel_end;
+	if(tmp < end)
+		start = tmp;
+	else {
+		start = end;
+		end = tmp;
+	}
+	end--;
+	sl = start/gh->cpl - gh->top_line;
+	el = end/gh->cpl - gh->top_line;
+
 	gdk_gc_set_foreground(gh->adisp_gc, &GTK_WIDGET(gh)->style->base[GTK_STATE_NORMAL]);
 	gdk_draw_rectangle(w->window, gh->adisp_gc, TRUE,
 					   0, imin*gh->char_height, w->allocation.width,
@@ -444,10 +550,39 @@ static void render_ascii_lines(GtkHex *gh, gint imin, gint imax) {
 		tmp = (gint)frm_len - (gint)((i - imin)*gh->cpl);
 		if(tmp <= 0)
 			return;
-		gdk_draw_text(w->window, gh->disp_font, gh->adisp_gc,
-					  0, gh->disp_font->ascent + i*gh->char_height,
-					  gh->disp_buffer + (i - imin)*gh->cpl,
-					  MIN(gh->cpl, tmp));
+
+		if(gh->sel_start != gh->sel_end) {
+			if(i == sl) {
+				cursor_off = start%gh->cpl;
+				if(i == el)
+					len = end - start + 1;
+				else
+					len = gh->cpl - cursor_off;
+				if(len > 0)
+					gtk_draw_box(GTK_WIDGET(gh)->style, gh->adisp->window,
+								 GTK_STATE_SELECTED, GTK_SHADOW_NONE,
+								 cursor_off*gh->char_width, i*gh->char_height,
+								 len*gh->char_width, gh->char_height);
+			}
+			else if(i == el) {
+				cursor_off = end%gh->cpl + 1;
+				if(cursor_off > 0)
+					gtk_draw_box(GTK_WIDGET(gh)->style, gh->adisp->window,
+								 GTK_STATE_SELECTED, GTK_SHADOW_NONE,
+								 0, i*gh->char_height,
+								 cursor_off*gh->char_width, gh->char_height);
+			}
+			else if(i > sl && i < el) {
+				gtk_draw_box(GTK_WIDGET(gh)->style, gh->adisp->window,
+							 GTK_STATE_SELECTED, GTK_SHADOW_NONE,
+							 0, i*gh->char_height,
+							 gh->cpl*gh->char_width, gh->char_height);
+			}
+		}  
+		/* Changes for Gnome 2.0 */
+		pango_layout_set_text (gh->alayout, gh->disp_buffer + (i - imin)*gh->cpl, MIN(gh->cpl, tmp));
+		gdk_draw_layout (w->window, gh->adisp_gc, 0, i*gh->char_height, gh->alayout);
+
 	}
 	
 	if((cursor_line >= imin) && (cursor_line <= imax) && (gh->cursor_shown))
@@ -479,9 +614,10 @@ static void render_offsets(GtkHex *gh, gint imin, gint imax) {
 	
 	for(i = imin; i <= imax; i++) {
 		sprintf(offstr, "%08X", (gh->top_line + i)*gh->cpl);
-		gdk_draw_text(w->window, gh->disp_font, gh->offsets_gc,
-					  0, gh->disp_font->ascent + i*gh->char_height,
-					  offstr, 8);
+		/* Changes for Gnome 2.0 */
+		pango_layout_set_text (gh->olayout, offstr, 8);
+		gdk_draw_layout (w->window, gh->offsets_gc, 0, i*gh->char_height, gh->olayout);
+
 	}
 }
 
@@ -711,42 +847,15 @@ static void display_scrolled(GtkAdjustment *adj, GtkHex *gh) {
 							 gh->offsets->allocation.width,
 							 source_max - source_min);
 		}
-
-		while ((event = gdk_event_get_graphics_expose (gh->xdisp->window)) != NULL) {
-			gtk_widget_event (gh->xdisp, event);
-			if (event->expose.count == 0) {
-				gdk_event_free (event);
-				break;
-			}
-			gdk_event_free (event);
-		}
-		
-		while ((event = gdk_event_get_graphics_expose (gh->adisp->window)) != NULL) {
-			gtk_widget_event (gh->adisp, event);
-			if (event->expose.count == 0) {
-				gdk_event_free (event);
-				break;
-			}
-			gdk_event_free (event);
-		}
-
-		if(gh->offsets) {
-			while ((event = gdk_event_get_graphics_expose (gh->offsets->window)) != NULL) {
-				gtk_widget_event (gh->offsets, event);
-				if (event->expose.count == 0) {
-					gdk_event_free (event);
-					break;
-				}
-				gdk_event_free (event);
-			}
-		}
 	}
-	
-	if (rect.height != 0) {
-		gtk_widget_draw(gh->xdisp, &rect);
-		gtk_widget_draw(gh->adisp, &rect);
-		if(gh->offsets)
-			gtk_widget_draw(gh->offsets, &rect);
+
+	rect.width = gh->xdisp->allocation.width;
+	gdk_window_invalidate_rect (gh->xdisp->window, &rect, FALSE);
+	rect.width = gh->adisp->allocation.width;
+	gdk_window_invalidate_rect (gh->adisp->window, &rect, FALSE);
+	if(gh->offsets) {
+		rect.width = gh->offsets->allocation.width;
+		gdk_window_invalidate_rect (gh->offsets->window, &rect, FALSE);
 	}
 }
 
@@ -758,7 +867,7 @@ static gboolean scroll_timeout_handler(GtkHex *gh) {
 		gtk_hex_set_cursor(gh, MAX(0, (int)(gh->cursor_pos - gh->cpl)));
 	else if(gh->scroll_dir > 0)
 		gtk_hex_set_cursor(gh, MIN(gh->document->file_size - 1,
-								   gh->cursor_pos + gh->cpl));    
+								   gh->cursor_pos + gh->cpl));
 	return TRUE;
 }
 
@@ -771,6 +880,7 @@ static void hex_button_cb(GtkWidget *w, GdkEventButton *event, GtkHex *gh) {
 			gh->scroll_timeout = -1;
 			gh->scroll_dir = 0;
 		}
+		gh->selecting = FALSE;
 		gtk_grab_remove(w);
 		gh->button = 0;
 	}
@@ -782,8 +892,14 @@ static void hex_button_cb(GtkWidget *w, GdkEventButton *event, GtkHex *gh) {
 
 		gh->button = event->button;
 		
-		if(gh->active_view == VIEW_HEX)
+		if(gh->active_view == VIEW_HEX) {
 			hex_to_pointer(gh, event->x, event->y);
+
+			if(!gh->selecting) {
+				gh->selecting = TRUE;
+				gtk_hex_set_selection(gh, gh->cursor_pos, gh->cursor_pos);
+			}
+		}
 		else {
 			hide_cursor(gh);
 			gh->active_view = VIEW_HEX;
@@ -834,6 +950,7 @@ static void ascii_button_cb(GtkWidget *w, GdkEventButton *event, GtkHex *gh) {
 			gh->scroll_timeout = -1;
 			gh->scroll_dir = 0;
 		}
+		gh->selecting = FALSE;
 		gtk_grab_remove(w);
 		gh->button = 0;
 	}
@@ -843,8 +960,13 @@ static void ascii_button_cb(GtkWidget *w, GdkEventButton *event, GtkHex *gh) {
 		
 		gtk_grab_add(w);
 		gh->button = event->button;
-		if(gh->active_view == VIEW_ASCII)
+		if(gh->active_view == VIEW_ASCII) {
 			ascii_to_pointer(gh, event->x, event->y);
+			if(!gh->selecting) {
+				gh->selecting = TRUE;
+				gtk_hex_set_selection(gh, gh->cursor_pos, gh->cursor_pos);
+			}
+		}
 		else {
 			hide_cursor(gh);
 			gh->active_view = VIEW_ASCII;
@@ -889,6 +1011,13 @@ static void ascii_motion_cb(GtkWidget *w, GdkEventMotion *event, GtkHex *gh) {
 
 static void show_offsets_widget(GtkHex *gh) {
 	gh->offsets = gtk_drawing_area_new();
+
+	/* Modify the font for the widget */
+	gtk_widget_modify_font (gh->offsets, gh->font_desc);
+ 
+	/* Create the pango layout for the widget */
+	gh->olayout = gtk_widget_create_pango_layout (gh->offsets, "");
+
 	gtk_widget_set_events (gh->offsets, GDK_EXPOSURE_MASK);
 	gtk_signal_connect(GTK_OBJECT(gh->offsets), "expose_event",
 					   GTK_SIGNAL_FUNC(offsets_expose), gh);
@@ -972,23 +1101,206 @@ static void gtk_hex_real_data_changed(GtkHex *gh, gpointer data) {
 	render_ascii_lines(gh, start_line, end_line);
 }
 
-static void gtk_hex_finalize(GtkObject *o) {
+static void bytes_changed(GtkHex *gh, gint start, gint end)
+{
+	gint start_line = start/gh->cpl - gh->top_line;
+	gint end_line = end/gh->cpl - gh->top_line;
+
+	if(end_line < 0 ||
+	   start_line > gh->vis_lines)
+		return;
+
+	start_line = MAX(start_line, 0);
+
+	render_hex_lines(gh, start_line, end_line);
+	render_ascii_lines(gh, start_line, end_line);
+}
+
+static void primary_get_cb(GtkClipboard *clipboard,
+						   GtkSelectionData *data, guint info,
+						   gpointer user_data) {
+	GtkHex *gh = GTK_HEX(user_data);
+	if(gh->sel_start != gh->sel_end) {
+		gint start_pos; 
+		gint end_pos;
+		guchar *text;
+		GtkHexClass *klass = GTK_HEX_CLASS(GTK_WIDGET_GET_CLASS(gh));
+
+		g_message("copy_primary");
+
+		start_pos = MIN(gh->sel_start, gh->sel_end);
+		end_pos = MAX(gh->sel_start, gh->sel_end);
+ 
+		text = hex_document_get_data(gh->document, start_pos,
+											 end_pos - start_pos);
+		gtk_selection_data_set_text(data, text, end_pos - start_pos);
+		g_free(text);
+	}
+}
+
+static void primary_clear_cb(GtkClipboard *clipboard,
+							 gpointer user_data_or_owner) {
+}
+
+static void gtk_hex_set_selection(GtkHex *gh, gint start, gint end)
+{
+	gint length = gh->document->file_size;
+	gint oe, os, ne, ns;
+	GtkHexClass *klass = GTK_HEX_CLASS(GTK_WIDGET_GET_CLASS(gh));
+
+	static const GtkTargetEntry targets[] = {
+		{ "STRING", 0, TARGET_STRING }
+	};
+	static const gint n_targets = sizeof(targets) / sizeof(targets[0]);
+
+	if (end < 0)
+		end = length;
+
+	if(gh->sel_start != gh->sel_end)
+		gtk_clipboard_clear(klass->primary);
+
+	os = MIN(gh->sel_start, gh->sel_end);
+	oe = MAX(gh->sel_start, gh->sel_end);
+	gh->sel_start = CLAMP(start, 0, length);
+	gh->sel_end = MIN(end, length);
+	ns = MIN(gh->sel_start, gh->sel_end);
+	ne = MAX(gh->sel_start, gh->sel_end);
+	if(ns != os && ne != oe) {
+		bytes_changed(gh, MIN(ns, os), MAX(ne, oe));
+	}
+	else if(ne != oe) {
+		bytes_changed(gh, MIN(ne, oe), MAX(ne, oe));
+	}
+	else if(ns != os) {
+		bytes_changed(gh, MIN(ns, os), MAX(ns, os));		
+	}
+
+	if(gh->sel_start != gh->sel_end)
+		gtk_clipboard_set_with_data(klass->primary, targets, n_targets,
+									primary_get_cb, primary_clear_cb,
+									gh);
+}
+
+void gtk_hex_select_region (GtkHex *gh,
+							gint         start,
+							gint         end) {
+	
+	gtk_hex_set_selection(gh, start, end);
+}
+
+void gtk_hex_delete_selection(GtkHex *gh)
+{
+	guint start;
+	guint end;
+
+	start = MIN(gh->sel_start, gh->sel_end);
+	end = MAX(gh->sel_end, gh->sel_start);
+
+	gtk_hex_select_region(gh, 0, 0);
+
+	if(start != end)
+		hex_document_delete_data(gh->document, MIN(start, end), end - start, TRUE);
+}
+
+void gtk_hex_copy_clipboard(GtkHex *gh)
+{
+	gtk_signal_emit_by_name(GTK_OBJECT(gh), "copy_clipboard");
+}
+
+void gtk_hex_cut_clipboard(GtkHex *gh)
+{
+	gtk_signal_emit_by_name(GTK_OBJECT(gh), "cut_clipboard");
+}
+
+void gtk_hex_paste_clipboard(GtkHex *gh)
+{
+	gtk_signal_emit_by_name(GTK_OBJECT(gh), "paste_clipboard");
+}
+
+static void gtk_hex_real_copy_clipboard(GtkHex *gh)
+{
+	guint32 time;
+	gint start_pos; 
+	gint end_pos;
+	GtkHexClass *klass = GTK_HEX_CLASS(GTK_WIDGET_GET_CLASS(gh));
+
+	g_message("copy_clipboard");
+	start_pos = MIN(gh->sel_start, gh->sel_end);
+	end_pos = MAX(gh->sel_start, gh->sel_end);
+ 
+	if(start_pos != end_pos) {
+		guchar *text = hex_document_get_data(gh->document, start_pos,
+											 end_pos - start_pos);
+		gtk_clipboard_set_text(klass->clipboard, text, end_pos - start_pos);
+		g_free(text);
+	}
+}
+
+static void gtk_hex_real_cut_clipboard(GtkHex *gh)
+{
+	g_message("cut_clipboard");
+
+	if(gh->sel_start != -1 && gh->sel_end != -1) {
+		gtk_hex_real_copy_clipboard(gh);
+		gtk_hex_delete_selection(gh);
+	}
+}
+
+static void gtk_hex_real_paste_clipboard(GtkHex *gh)
+{
+	GtkHexClass *klass = GTK_HEX_CLASS(GTK_WIDGET_GET_CLASS(gh));
+	gchar *text;
+
+	g_message("paste_clipboard");
+
+	text = gtk_clipboard_wait_for_text(klass->clipboard);
+	if(text) {
+		hex_document_set_data(gh->document, gh->cursor_pos,
+							  strlen(text), 0, text, TRUE);
+		g_free(text);
+	}
+}
+
+static void gtk_hex_finalize(GObject *o) {
 	GtkHex *gh = GTK_HEX(o);
 	
 	if(gh->disp_buffer)
 		g_free(gh->disp_buffer);
+
+	if (gh->disp_font_metrics)
+		pango_font_metrics_unref (gh->disp_font_metrics);
+
+	if (gh->font_desc)
+		pango_font_description_free (gh->font_desc);
+
+	if (gh->xlayout)
+		g_object_unref (G_OBJECT (gh->xlayout));
 	
-	if(GTK_OBJECT_CLASS(parent_class)->finalize)
-		(* GTK_OBJECT_CLASS(parent_class)->finalize)(o);  
+	if (gh->alayout)
+		g_object_unref (G_OBJECT (gh->alayout));
+	
+	if (gh->olayout)
+		g_object_unref (G_OBJECT (gh->olayout));
+	
+	/* Changes for Gnome 2.0 -- SnM */	
+	if(G_OBJECT_CLASS(parent_class)->finalize)
+		(* G_OBJECT_CLASS(parent_class)->finalize)(G_OBJECT(o));  
 }
 
 static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 	GtkHex *gh = GTK_HEX(w);
 	guint old_cp = gh->cursor_pos;
 	gint ret = TRUE;
+	gint cx, cy;
 
 	hide_cursor(gh);
 
+	if(!(event->state & GDK_SHIFT_MASK)) {
+		gh->selecting = FALSE;
+	}
+	else {
+		gh->selecting = TRUE;
+	}
 	switch(event->keyval) {
 	case GDK_BackSpace:
 		if(gh->cursor_pos > 0) {
@@ -997,6 +1309,21 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 			gtk_hex_set_cursor(gh, gh->cursor_pos - 1);
 		}
 		break;
+	case GDK_Tab:
+	case GDK_KP_Tab:
+			if (event->state & GDK_CONTROL_MASK) {
+				if (gh->active_view == VIEW_ASCII) {
+					get_acoords(gh, gh->cursor_pos, &cx, &cy);
+					gtk_draw_box(GTK_WIDGET(gh)->style, gh->adisp->window, GTK_STATE_NORMAL, GTK_SHADOW_IN, cx, cy, gh->char_width, gh->char_height-1);
+					gh->active_view = VIEW_HEX;
+				}
+				else {
+					get_xcoords(gh, gh->cursor_pos, &cx, &cy);
+					gtk_draw_box(GTK_WIDGET(gh)->style, gh->xdisp->window, GTK_STATE_NORMAL, GTK_SHADOW_IN, cx, cy, gh->char_width, gh->char_height-1);
+					gh->active_view = VIEW_ASCII;
+				}
+			}
+			break;
 	case GDK_Delete:
 		if(gh->cursor_pos < gh->document->file_size) {
 			hex_document_set_data(gh->document, gh->cursor_pos,
@@ -1026,17 +1353,27 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 			switch(event->keyval) {
 			case GDK_Left:
 			case GDK_KP_4:
-				gh->lower_nibble = !gh->lower_nibble;
-				if(gh->lower_nibble)
+				if(!(event->state & GDK_SHIFT_MASK)) {
+					gh->lower_nibble = !gh->lower_nibble;
+					if(gh->lower_nibble)
+						gtk_hex_set_cursor(gh, gh->cursor_pos - 1);
+				}
+				else {
 					gtk_hex_set_cursor(gh, gh->cursor_pos - 1);
+				}
 				break;
 			case GDK_Right:
 			case GDK_KP_6:
 				if(gh->cursor_pos >= gh->document->file_size)
 					break;
-				gh->lower_nibble = !gh->lower_nibble;
-				if(!gh->lower_nibble)
+				if(!(event->state & GDK_SHIFT_MASK)) {
+					gh->lower_nibble = !gh->lower_nibble;
+					if(!gh->lower_nibble)
+						gtk_hex_set_cursor(gh, gh->cursor_pos + 1);
+				}
+				else {
 					gtk_hex_set_cursor(gh, gh->cursor_pos + 1);
+				}
 				break;
 			default:
 				if(event->length != 1)
@@ -1153,13 +1490,6 @@ static void gtk_hex_size_allocate(GtkWidget *w, GtkAllocation *alloc) {
 	show_cursor(gh);
 }
 
-static void gtk_hex_draw(GtkWidget *w, GdkRectangle *area) {
-	if(GTK_WIDGET_CLASS(parent_class)->draw)
-		(* GTK_WIDGET_CLASS(parent_class)->draw)(w, area);
-
-	draw_shadow(w, area);
-}
-
 static gint gtk_hex_expose(GtkWidget *w, GdkEventExpose *event) {
 	draw_shadow(w, &event->area);
 	
@@ -1183,40 +1513,76 @@ static void gtk_hex_size_request(GtkWidget *w, GtkRequisition *req) {
 		2*GTK_CONTAINER(w)->border_width;
 }
 
-static void gtk_hex_class_init(GtkHexClass *klass) {
-	GtkObjectClass *object_class;
+static void gtk_hex_class_init(GtkHexClass *klass, gpointer data) {
+	GtkObjectClass *object_class = GTK_OBJECT_CLASS(klass);
+	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
 
-	object_class = (GtkObjectClass*) klass;
-	
-	gtkhex_signals[CURSOR_MOVED_SIGNAL] = gtk_signal_new ("cursor_moved",
-														  GTK_RUN_FIRST,
-														  object_class->type,
-														  GTK_SIGNAL_OFFSET (GtkHexClass, cursor_moved),
-														  gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
-	gtkhex_signals[DATA_CHANGED_SIGNAL] = gtk_signal_new ("data_changed",
-														  GTK_RUN_FIRST,
-														  object_class->type,
-														  GTK_SIGNAL_OFFSET (GtkHexClass, data_changed),
-														  gtk_hex_marshaller, GTK_TYPE_NONE, 2,
-														  GTK_TYPE_INT, GTK_TYPE_INT);
-	
-	gtk_object_class_add_signals (object_class, gtkhex_signals, LAST_SIGNAL);
+	parent_class = g_type_class_peek_parent(klass);
 	
+	gtkhex_signals[CURSOR_MOVED_SIGNAL] = 
+		gtk_signal_new ("cursor_moved",
+			GTK_RUN_FIRST,
+			GTK_CLASS_TYPE(object_class),
+			GTK_SIGNAL_OFFSET (GtkHexClass, cursor_moved),
+			gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
+
+	gtkhex_signals[DATA_CHANGED_SIGNAL] = 
+		gtk_signal_new ("data_changed",
+			GTK_RUN_FIRST,
+			GTK_CLASS_TYPE(object_class),
+			GTK_SIGNAL_OFFSET (GtkHexClass, data_changed),
+			ghex_marshal_VOID__POINTER, GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);
+
+	gtkhex_signals[CUT_CLIPBOARD_SIGNAL] = 
+		gtk_signal_new ("cut_clipboard",
+			GTK_RUN_FIRST,
+			GTK_CLASS_TYPE(object_class),
+			GTK_SIGNAL_OFFSET (GtkHexClass, cut_clipboard),
+			gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
+
+	gtkhex_signals[COPY_CLIPBOARD_SIGNAL] = 
+		gtk_signal_new ("copy_clipboard",
+			GTK_RUN_FIRST,
+			GTK_CLASS_TYPE(object_class),
+			GTK_SIGNAL_OFFSET (GtkHexClass, copy_clipboard),
+			gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
+
+	gtkhex_signals[PASTE_CLIPBOARD_SIGNAL] = 
+		gtk_signal_new ("paste_clipboard",
+			GTK_RUN_FIRST,
+			GTK_CLASS_TYPE(object_class),
+			GTK_SIGNAL_OFFSET (GtkHexClass, paste_clipboard),
+			gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
+
 	klass->cursor_moved = NULL;
 	klass->data_changed = gtk_hex_real_data_changed;
-	
-	GTK_WIDGET_CLASS(klass)->draw = gtk_hex_draw;
+	klass->cut_clipboard = gtk_hex_real_cut_clipboard;
+	klass->copy_clipboard = gtk_hex_real_copy_clipboard;
+	klass->paste_clipboard = gtk_hex_real_paste_clipboard;
+
+	klass->primary = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
+	klass->clipboard = gtk_clipboard_get(GDK_NONE);
+
 	GTK_WIDGET_CLASS(klass)->size_allocate = gtk_hex_size_allocate;
 	GTK_WIDGET_CLASS(klass)->size_request = gtk_hex_size_request;
 	GTK_WIDGET_CLASS(klass)->expose_event = gtk_hex_expose;
 	GTK_WIDGET_CLASS(klass)->key_press_event = gtk_hex_key_press;
 	GTK_WIDGET_CLASS(klass)->realize = gtk_hex_realize;
-	GTK_OBJECT_CLASS(klass)->finalize = gtk_hex_finalize;
+
+	/* Changed in Gnome 2.0 -- SnM */
+	G_OBJECT_CLASS(klass)->finalize = gtk_hex_finalize;
+
+	klass->cmap = gdk_colormap_get_system();
 	
+	memcpy(&klass->cl_cursor_p, &cl_cursor_p, sizeof(GdkColor));
+	memcpy(&klass->cl_cursor_a, &cl_cursor_a, sizeof(GdkColor));
+	gdk_colormap_alloc_color(klass->cmap, &klass->cl_cursor_p, FALSE, TRUE);
+	gdk_colormap_alloc_color(klass->cmap, &klass->cl_cursor_a, FALSE, TRUE);
+
 	parent_class = gtk_type_class (gtk_fixed_get_type ());
 }
 
-static void gtk_hex_init(GtkHex *gh) {
+static void gtk_hex_init(GtkHex *gh, gpointer klass) {
 	gh->scroll_timeout = -1;
 
 	gh->disp_buffer = NULL;
@@ -1232,19 +1598,29 @@ static void gtk_hex_init(GtkHex *gh) {
 	gh->cursor_shown = FALSE;
 	gh->button = 0;
 	gh->insert = FALSE; /* default to overwrite mode */
+	gh->selecting = FALSE;
 
 	/* get ourselves a decent monospaced font for rendering text */
-	gh->disp_font = gdk_font_load(DEFAULT_FONT);
-	gh->char_width = get_max_char_width(gh->disp_font);
-	gh->char_height = gh->disp_font->ascent + gh->disp_font->descent + 2;
+	gh->disp_font_metrics = gtk_hex_load_font (DEFAULT_FONT);
+	gh->font_desc = pango_font_description_from_string (DEFAULT_FONT);
+
+	gh->char_width = get_max_char_width(gh, gh->disp_font_metrics);
+	gh->char_height = PANGO_PIXELS (pango_font_metrics_get_ascent (gh->disp_font_metrics)) + PANGO_PIXELS (pango_font_metrics_get_descent (gh->disp_font_metrics)) + 2;
+
 	
 	GTK_WIDGET_SET_FLAGS(gh, GTK_CAN_FOCUS);
 	gtk_widget_set_events(GTK_WIDGET(gh), GDK_KEY_PRESS_MASK);
 	gtk_container_border_width(GTK_CONTAINER(gh), DISPLAY_BORDER);
 	
 	gh->adj = GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, 0.0, 0.0, 0.0, 0.0));
-	
 	gh->xdisp = gtk_drawing_area_new();
+
+	/* Modify the font for the widget */
+	gtk_widget_modify_font (gh->xdisp, gh->font_desc);
+
+	/* Create the pango layout for the widget */
+	gh->xlayout = gtk_widget_create_pango_layout (gh->xdisp, "");
+
 	gtk_widget_set_events (gh->xdisp, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
 						   GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_MOTION_MASK);
 	gtk_signal_connect(GTK_OBJECT(gh->xdisp), "expose_event",
@@ -1261,6 +1637,13 @@ static void gtk_hex_init(GtkHex *gh) {
 	gtk_widget_show(gh->xdisp);
 	
 	gh->adisp = gtk_drawing_area_new();
+
+	/* Modify the font for the widget */
+	gtk_widget_modify_font (gh->adisp, gh->font_desc);
+
+	/* Create the pango layout for the widget */
+	gh->alayout = gtk_widget_create_pango_layout (gh->adisp, "");
+
 	gtk_widget_set_events (gh->adisp, GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
 						   GDK_BUTTON_RELEASE_MASK |GDK_BUTTON_MOTION_MASK);
 	gtk_signal_connect(GTK_OBJECT(gh->adisp), "expose_event",
@@ -1288,23 +1671,33 @@ guint gtk_hex_get_type() {
 	static guint gh_type = 0;
 	
 	if(!gh_type) {
-		GtkTypeInfo gh_info = {
-			"GtkHex", sizeof(GtkHex), sizeof(GtkHexClass),
-			(GtkClassInitFunc) gtk_hex_class_init,
-			(GtkObjectInitFunc) gtk_hex_init,
-			(GtkArgSetFunc) NULL,
-			(GtkArgGetFunc) NULL,
+		GTypeInfo gh_info = {
+			sizeof (GtkHexClass),
+			NULL,		/* base_init */
+			NULL,		/* base_finalize */
+			(GClassInitFunc) gtk_hex_class_init,
+			NULL,		/* class_finalize */
+			NULL,		/* class_data */
+			sizeof (GtkHex),
+			0,
+			(GInstanceInitFunc) gtk_hex_init	
 		};
-		
-		gh_type = gtk_type_unique(gtk_fixed_get_type(), &gh_info);
+	
+		gh_type = g_type_register_static (gtk_fixed_get_type(),
+							"GtkHex",
+							&gh_info,
+							0);	
 	}
 	
 	return gh_type;
 }
 
 GtkWidget *gtk_hex_new(HexDocument *owner) {
-	GtkHex *gh = gtk_type_new(gtk_hex_get_type());
-	
+	GtkHex *gh;
+
+	gh = GTK_HEX (g_object_new (gtk_hex_get_type(), NULL));
+	g_return_val_if_fail (gh != NULL, NULL);
+
 	gh->document = owner;
 	
 	return GTK_WIDGET(gh);
@@ -1321,9 +1714,22 @@ void gtk_hex_set_nibble(GtkHex *gh, gint lower_nibble) {
 	g_return_if_fail(gh != NULL);
 	g_return_if_fail(GTK_IS_HEX(gh));
 
-	hide_cursor(gh);
-	gh->lower_nibble = lower_nibble;
-	show_cursor(gh);
+	if(gh->selecting) {
+		bytes_changed(gh, gh->cursor_pos, gh->cursor_pos);
+		gh->lower_nibble = lower_nibble;
+	}
+	else if(gh->sel_end != gh->sel_start) {
+		guint start = MIN(gh->sel_start, gh->sel_end);
+		guint end = MAX(gh->sel_start, gh->sel_end);
+		gh->sel_end = gh->sel_start = 0;
+		bytes_changed(gh, start, end);
+		gh->lower_nibble = lower_nibble;
+	}
+	else {
+		hide_cursor(gh);
+		gh->lower_nibble = lower_nibble;
+		show_cursor(gh);
+	}
 }
 
 /*
@@ -1331,7 +1737,8 @@ void gtk_hex_set_nibble(GtkHex *gh, gint lower_nibble) {
  */
 void gtk_hex_set_cursor(GtkHex *gh, gint index) {
 	guint y;
-	
+	guint old_pos = gh->cursor_pos;
+
 	g_return_if_fail(gh != NULL);
 	g_return_if_fail(GTK_IS_HEX(gh));
 
@@ -1342,7 +1749,7 @@ void gtk_hex_set_cursor(GtkHex *gh, gint index) {
 		hide_cursor(gh);
 		
 		gh->cursor_pos = index;
-		
+
 		if(gh->cpl == 0)
 			return;
 		
@@ -1361,7 +1768,17 @@ void gtk_hex_set_cursor(GtkHex *gh, gint index) {
 			gh->lower_nibble = FALSE;
 		
 		gtk_signal_emit_by_name(GTK_OBJECT(gh), "cursor_moved");
-		
+
+		if(gh->selecting) {
+			gtk_hex_set_selection(gh, gh->sel_start, gh->cursor_pos);
+			bytes_changed(gh, MIN(gh->cursor_pos, old_pos), MAX(gh->cursor_pos, old_pos));
+		}
+		else if(gh->sel_end != gh->sel_start) {
+			guint start = MIN(gh->sel_start, gh->sel_end);
+			guint end = MAX(gh->sel_start, gh->sel_end);
+			gh->sel_end = gh->sel_start = 0;
+			bytes_changed(gh, start, end);
+		}
 		show_cursor(gh);
 	}
 }
@@ -1371,6 +1788,7 @@ void gtk_hex_set_cursor(GtkHex *gh, gint index) {
  */
 void gtk_hex_set_cursor_xy(GtkHex *gh, gint x, gint y) {
 	gint cp;
+	guint old_pos = gh->cursor_pos;
 
 	g_return_if_fail(gh != NULL);
 	g_return_if_fail(GTK_IS_HEX(gh));
@@ -1398,6 +1816,17 @@ void gtk_hex_set_cursor_xy(GtkHex *gh, gint x, gint y) {
 		
 		gtk_signal_emit_by_name(GTK_OBJECT(gh), "cursor_moved");
 		
+		if(gh->selecting) {
+			gtk_hex_set_selection(gh, gh->sel_start, gh->cursor_pos);
+			bytes_changed(gh, MIN(gh->cursor_pos, old_pos), MAX(gh->cursor_pos, old_pos));
+		}
+		else if(gh->sel_end != gh->sel_start) {
+			guint start = MIN(gh->sel_start, gh->sel_end);
+			guint end = MAX(gh->sel_start, gh->sel_end);
+			gh->sel_end = gh->sel_start = 0;
+			bytes_changed(gh, start, end);
+		}
+
 		show_cursor(gh);
 	}
 }
@@ -1434,6 +1863,7 @@ void gtk_hex_set_group_type(GtkHex *gh, guint gt) {
 
 	hide_cursor(gh);
 	gh->group_type = gt;
+	recalc_displays(gh, GTK_WIDGET(gh)->allocation.width, GTK_WIDGET(gh)->allocation.height);
 	gtk_widget_queue_resize(GTK_WIDGET(gh));
 	show_cursor(gh);
 }
@@ -1441,17 +1871,32 @@ void gtk_hex_set_group_type(GtkHex *gh, guint gt) {
 /*
  * sets font for displaying data
  */
-void gtk_hex_set_font(GtkHex *gh, GdkFont *font) {
+void gtk_hex_set_font(GtkHex *gh, PangoFontMetrics *font_metrics, PangoFontDescription *font_desc) {
+	PangoFontDescription *tmp_desc;
 	g_return_if_fail(gh != NULL);
 	g_return_if_fail(GTK_IS_HEX(gh));
 
-	if(gh->disp_font)
-		gdk_font_unref(gh->disp_font);
-	
-	gh->disp_font = gdk_font_ref(font);
-	gh->char_width = get_max_char_width(gh->disp_font);
-	gh->char_height = gh->disp_font->ascent + gh->disp_font->descent + 2;
+	if (gh->disp_font_metrics)
+		pango_font_metrics_unref (gh->disp_font_metrics);
+
+	if (gh->font_desc)
+		pango_font_description_free (gh->font_desc);
 	
+	gh->disp_font_metrics = pango_font_metrics_ref (font_metrics);
+	gh->font_desc = pango_font_description_copy (font_desc);
+
+	if (gh->xdisp)
+		gtk_widget_modify_font (gh->xdisp, gh->font_desc);
+
+	if (gh->adisp)
+		gtk_widget_modify_font (gh->adisp, gh->font_desc);
+
+	if (gh->offsets)
+		gtk_widget_modify_font (gh->offsets, gh->font_desc);
+
+
+	gh->char_width = get_max_char_width(gh, gh->disp_font_metrics);
+	gh->char_height = PANGO_PIXELS (pango_font_metrics_get_ascent (gh->disp_font_metrics)) + PANGO_PIXELS (pango_font_metrics_get_descent (gh->disp_font_metrics)) + 2;
 	recalc_displays(gh, GTK_WIDGET(gh)->allocation.width, GTK_WIDGET(gh)->allocation.height);
 	
 	redraw_widget(GTK_WIDGET(gh));
@@ -1486,3 +1931,62 @@ void gtk_hex_set_insert_mode(GtkHex *gh, gboolean insert)
 		gh->cursor_pos = gh->document->file_size - 1;
 }
 
+PangoFontMetrics* gtk_hex_load_font (const char *font_name)
+{
+	PangoContext *context;
+	PangoFont *new_font;
+	PangoFontDescription *new_desc;
+	PangoFontMetrics *new_metrics;
+
+	new_desc = pango_font_description_from_string (font_name);
+
+	context = gdk_pango_context_get();
+
+	/* FIXME - Should get the locale language here */
+	pango_context_set_language (context, gtk_get_default_language());
+
+	new_font = pango_context_load_font (context, new_desc);
+
+	new_metrics = pango_font_get_metrics (new_font, pango_context_get_language (context));
+
+	pango_font_description_free (new_desc);
+	g_object_unref (G_OBJECT (context));
+	g_object_unref (G_OBJECT (new_font));
+
+	return new_metrics;
+}
+
+void add_atk_namedesc (GtkWidget *widget, const gchar *name, const gchar *desc)
+{
+        AtkObject *atk_widget;
+
+        g_return_if_fail (GTK_IS_WIDGET (widget));
+        atk_widget = gtk_widget_get_accessible (widget);
+
+        if (name)
+                atk_object_set_name (atk_widget, name);
+        if (desc)
+                atk_object_set_description (atk_widget, desc);
+}
+
+void add_atk_relation (GtkWidget *obj1, GtkWidget *obj2, AtkRelationType type)
+{
+
+        AtkObject *atk_obj1, *atk_obj2;
+        AtkRelationSet *relation_set;
+        AtkRelation *relation;
+
+        g_return_if_fail (GTK_IS_WIDGET (obj1));
+        g_return_if_fail (GTK_IS_WIDGET (obj2));
+
+        atk_obj1 = gtk_widget_get_accessible (obj1);
+        atk_obj2 = gtk_widget_get_accessible (obj2);
+
+        relation_set = atk_object_ref_relation_set (atk_obj1);
+        relation = atk_relation_new (&atk_obj2, 1, type);
+        atk_relation_set_add (relation_set, relation);
+        g_object_unref (G_OBJECT (relation));
+
+}
+
+
